"""DataIntegration Task Runner.

This Module can run data-integration tasks, HTTP/RESTful calls
(or even any arbitrary python function),
received parametriced tasks and return the results.

Task Runner receives different command-line arguments:

--variables: override variables passed between components/processes.
--attributes: override root-level attributes of components.
--conditions: override values in components with conditions.
--params: any value in params ins passed as kwargs
--args: list of arguments to be used by some functions (like replacement of values)

* For Calling a task:
 --program=program_name --task=task_name --params
* For calling Python Functions:
 --function=path.to.python.import --params
* System Command:
 --command=path.to.command --params
* or even Calling a URL:
 --url=URI --params
"""
from typing import Any
from collections.abc import Callable
import uuid
import random
import asyncio
import signal
import uvloop
from asyncdb import AsyncDB
from navconfig.logging import logging
# Worker:
from qw.wrappers import TaskWrapper
from qw.client import QClient
# DataIntegration
from flowtask.conf import (
    MEMCACHE_HOST,
    MEMCACHE_PORT,
    WORKER_LIST
)
from flowtask.utils.stats import TaskMonitor
from flowtask.tasks.task import Task
from flowtask.tasks.command import Command
from flowtask.parsers.argparser import ConfigParser
from flowtask.exceptions import (
    TaskError,
    TaskException,
    NotSupported
)


def my_handler():
    print('Stopping')
    for task in asyncio.all_tasks():
        task.cancel()

# os.environ['PYTHONASYNCIODEBUG'] = '1'
asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
uvloop.install()

# Worker definition

QW = QClient(
    worker_list=WORKER_LIST
)

class TaskRunner:
    """
    TaskRunner.

        Execution of DataIntegration Tasks.
    """
    logger: logging.Logger = None

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop = None,
        worker=None,
        **kwargs
    ):
        """Init Method for Task Runner."""
        self._task = None
        self._command: str = None
        self._memory: Callable = None
        self._fn: str = None
        self._url: str = None
        self._taskname = ''
        self.task_type: str = 'task'
        self.stat: TaskMonitor = None
        # arguments passed by command line:
        self._conditions = {}
        self._attrs = {}
        self.ignore_steps = []
        self.run_only = []
        self._stepattrs = {}
        self._kwargs = {}
        self._variables = {}
        self._result: Any = None
        if loop:
            self._loop = loop
        else:
            self._loop = asyncio.new_event_loop()
        self._loop.add_signal_handler(signal.SIGINT, my_handler)
        # Task ID:
        self.task_id = uuid.uuid1(node=random.getrandbits(48) | 0x010000000000)
        # argument Parser
        if 'new_args' in kwargs:
            new_args = kwargs['new_args']
            del kwargs['new_args']
        else:
            new_args = []
        parser = ConfigParser()
        parser.parse(new_args)
        self._argparser = parser
        self._options = parser.options
        # program definition
        self._program = self._options.program
        try:
            self._program = kwargs['program']
            del kwargs['program']
        except KeyError:
            pass
        if not self._program:
            self._program = 'navigator'
        # DEBUG
        try:
            self._debug = kwargs['debug']
            del kwargs['debug']
        except KeyError:
            self._debug = self._options.debug
        # logger
        self.logger = logging.getLogger('FlowTask')
        if self._debug:
            self.logger.setLevel(logging.DEBUG)
        self.logger.notice(f'Program is: > {self._program}')
        # about Worker Information:
        self.worker = worker
        ## Task Storage:
        self._taskstorage = self._options.storage
        try:
            self.no_worker = self._options.no_worker
        except (ValueError, TypeError):
            self.no_worker = False
        if kwargs:
            # remain args go to kwargs:
            self._kwargs = {**kwargs}


    @property
    def stats(self) -> TaskMonitor:
        """stats.
        Return a TaskMonitor object with all collected stats.
        Returns:
            TaskMonitor: stat object.
        """
        return self.stat

    def options(self) -> dict:
        return self._options

    async def __aenter__(self) -> "TaskRunner":
        """ Magic Context Methods """
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """ Magic Context Methods """
        # clean up anything you need to clean up
        await self.close()

    @property
    def result(self):
        return self._result

    async def setup_memcached(self):
        # create a Memcache connection
        params = {
            "host": MEMCACHE_HOST,
            "port": MEMCACHE_PORT
        }
        try:
            self._memory = AsyncDB('memcache', params=params)
            await self._memory.connection()
            connected = self._memory.is_connected()
            logging.debug(f'DI: Memcached Connected: {connected}')
        except Exception as err:  # pylint: disable=W0718
            self._memory = None
            logging.exception(err)

    async def start(self):
        # setting up a memcached connection:
        await self.setup_memcached()
        # define the type of task:
        if self._options.command:
            self._command = self._options.command
            self.task_type = 'command'
            self._taskname = f"{self._command!s}"
        elif self._options.function:
            self._fn = self._options.function
            self.task_type = 'function'
            self._taskname = f"{self._fn!s}"
        elif self._options.task:
            self._taskname = self._options.task
            self.task_type = 'task'
        else:
            raise NotSupported(
                "Task Runner: Unknown or missing Task type."
            )
        # initialize the Task Monitor
        try:
            self.stat = TaskMonitor(
                name=self._taskname,
                program=self._program,
                task_id=self.task_id
            )
            await self.stat.start()
        except Exception as err:
            raise TaskError(
                f"Task Runner: Error on TaskMonitor: {err}"
            ) from err
        # create the task object:
        if self.task_type == 'command':
            try:
                self._task = Command(
                    task_id=self.task_id,
                    task=self._command,
                    program=self._program,
                    loop=self._loop,
                    memory=self._memory,
                    parser=self._argparser,
                    **self._kwargs
                )
            except Exception as err:
                raise TaskException(err) from err
        elif self.task_type == 'task':
            try:
                self._task = Task(
                    task_id=self.task_id,
                    task=self._taskname,
                    program=self._program,
                    loop=self._loop,
                    memory=self._memory,
                    parser=self._argparser,
                    storage=self._taskstorage,
                    **self._kwargs
                )
            except Exception as err:
                raise TaskException(err) from err
        # then, try to "start" the task:
        logging.debug(f'::::: DI: Running {self.task_type}: {self._task!r}')
        try:
            self._task.setStat(self.stat)
            await self._task.start()
        except Exception as err:
            logging.error(err)
            raise TaskException(str(err)) from err
        return True

    async def run(self):
        try:
            if self.no_worker:
                try:
                    self._result = await self._task.run()
                finally:
                    await self._task.close()
            else:
                # sent task to Worker
                task = TaskWrapper(
                    program=self._program,
                    task_id=self.task_id,
                    task=self._taskname,
                    debug=self._debug,
                    parser=self._argparser,
                    **self._kwargs
                )
                if self._options.queued is True:
                    task.queued = True
                    result = await QW.queue(task)
                else:
                    task.queued = False
                    result = await QW.run(task)
                if isinstance(result, BaseException):
                    raise result()
                else:
                    self._result = result
        except Exception as err:
            logging.error(err)
            raise

        return True

    async def close(self):
        if self._debug:
            logging.debug(f'ENDING TASK {self._taskname}')
        try:
            await self.stat.stop()
        except Exception as err:  # pylint: disable=W0718
            logging.exception(err)
        try:
            await self._memory.close()
        except Exception as err:  # pylint: disable=W0718
            logging.exception(err)
